#include <Servo.h>

#include <ESP8266WiFi.h>  //Wifi
#include <PubSubClient.h> //MQTT

#define MIN(a,b)  a < b ? a : b
#define MAX(a,b)  a > b ? a : b

/* variables to manage the servos from the robotic arm */
Servo servoD1;
int servoD1_min, servoD1_max;
Servo servoD2;
int servoD2_min, servoD2_max;
Servo servoD3;
int servoD3_min, servoD3_max;
Servo servoD4;
int servoD4_min, servoD4_max;
int init_value = 90;
int delay_servo = 100;

int speed_move = 10;
int delay_t = 500;

int catch_object = 0;

//Wifi variables
const char* ssid = "FabAcademy";
const char* password = "fabacademy";
IPAddress wemosIP; //IP addres that wemos received

//server to receive request from app
WiFiServer httpServer(80);
int counter = 0;

//MQTT variables
const char* mqtt_server = "192.168.4.1";
const int mqtt_port = 1883;
WiFiClient espClient;
char* mqttTopicCatcherArm = "MQTTTriggerCatcherArm";
PubSubClient MQTTclient(espClient);

unsigned long lastMsg = 0;
#define MSG_BUFFER_SIZE  (50)
char msgReceived[MSG_BUFFER_SIZE];

void move_nicely_servo(Servo servo, int pos){
  int ini = servo.read();
  int steps_move = ini > pos ? -1 : 1;
  while(ini != pos){
    ini += steps_move;
    servo.write(ini);
    delay(delay_servo);
  }
}

void move_servo_clock(Servo servo, int max_v){
  int pos = servo.read();
  int steps_move = MIN(max_v, pos+speed_move);
  move_nicely_servo(servo, steps_move);
}
void move_servo_counterclock(Servo servo, int min_v){
  int pos = servo.read();
  int steps_move = MAX(min_v, pos-speed_move);
  move_nicely_servo(servo, pos-speed_move);
}

void setup_servos(){
  Serial.println("Preparing servos...");
  /* Settings for servo that moves UpDown (UDL UDR)*/
  servoD1.attach(D1);
  servoD1_min = 60;
  servoD1_max = 130;
  /* Settings for servo that moves FrontBack (FBL FBR)*/
  servoD2.attach(D2);
  servoD2_min = 60;
  servoD2_max = 110;
  /* Settings for servo that moves Wrist (WL WR)*/
  servoD3.attach(D3);
  servoD3_min = 50;
  servoD3_max = 140;
  /* Settings for servo that Catches (HOLD UNHOLD)*/
  servoD4.attach(D4);
  servoD4_min = 75;
  servoD4_max = 110;
  
  /* Set the initial position */
  Serial.println("Initial position...");
  move_nicely_servo(servoD1, init_value);
  move_nicely_servo(servoD2, init_value);
  move_nicely_servo(servoD3, init_value);
  move_nicely_servo(servoD4, init_value);
  
  Serial.println("Servos ready.");
}

void setup_wifi() {
  delay(10);
  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.println(ssid);

  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.print(".");
  }

  randomSeed(micros());

  wemosIP = WiFi.localIP();
  Serial.println("");
  Serial.println("WiFi connected");
  Serial.print("IP address: ");
  Serial.println(wemosIP);
}

void setup_httpServer(){
  /* Initialize the server */
  httpServer.begin();
  Serial.println("Server started...");
 
  // Print the IP address
  Serial.print("Use this URL to connect: ");
  Serial.print("http://");
  Serial.print(WiFi.localIP());
  Serial.println("/");
}

void handleClientRequestServer(){
  // Check if a client has connected
  WiFiClient client = httpServer.available();
  if (!client) {
    return;
  }

  // Wait until the client sends some data
  Serial.println("New client");
  while(!client.available()){
    Serial.print(".");
    delay(1000);
  }
  String request = client.readStringUntil('\r');

  Serial.print("Value read from request: ");
  Serial.println(request);
  client.flush();

  String ret;
  //servoD1_min   servoD1_max
  if (request.indexOf("move/UP") != -1)  {
    ret = "Move base arm left...";
    move_servo_clock(servoD1, servoD1_max);
  }
  else if (request.indexOf("move/DOWN") != -1)  {
    ret = "Move base arm right...";
    move_servo_counterclock(servoD1, servoD1_min);
  }
  else if (request.indexOf("/move/FRONT") != -1)  {
    ret = "Move UpDown arm left...";
    move_servo_clock(servoD2, servoD2_max);
  }
  else if (request.indexOf("/move/BACK") != -1)  {
    ret = "Move UpDown arm right...";
    move_servo_counterclock(servoD2, servoD2_min);
  }
  else if (request.indexOf("/move/WristUP") != -1)  {
    ret = "Move FrontBack arm left...";
    move_servo_clock(servoD3, servoD3_max);
  }
  else if (request.indexOf("/move/WristDOWN") != -1)  {
    ret = "Move FrontBack arm right...";
    move_servo_counterclock(servoD3, servoD3_min);
  }
  else if (request.indexOf("HOLD") != -1)  {
    ret = "Hold object...";
    move_nicely_servo(servoD4, servoD4_max);
  }
  else if (request.indexOf("UNHOLD") != -1)  {
    ret = "Unhold object...";
    move_nicely_servo(servoD4, servoD4_min);
  }
  else {
    ret = "Unknown request [" + request + "]";
  }
  counter++;
 
  // Return the response
  client.println("HTTP/1.1 200 OK");
  client.println("Content-Type: text/html");
  client.println(""); //  do not forget this one
  //client.println("<!DOCTYPE HTML>");
  //client.println("<html>");
  client.print("Counter: ");
  client.println(counter);
  client.print("Last move: ");
  client.println(ret);
  //client.println("</html>");
 
  delay(delay_t);
  Serial.println("Client disonnected");
}

void mqttMessageCallback(char* topic, byte* payload, unsigned int length) {
  for(int i=0; i<length;++i){
    msgReceived[i] = (char)payload[i];
  }
  msgReceived[length] = '\0';
  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.print("]: ");
  Serial.println(msgReceived);
  
  catch_object = 1;
}

void setupMqttClient() {
  // to call in the setup of the arduino
  MQTTclient.setServer(mqtt_server, mqtt_port);
  MQTTclient.setCallback(mqttMessageCallback);
  mqttConnect();
}

void mqttSubscribeTopic(char* topic){
  MQTTclient.subscribe(topic);
  Serial.print("MQTT subscribed to: ");
  Serial.println(topic);
}

void mqttConnect() {
  // Loop until MQTT is reconnected
  while (!MQTTclient.connected()) {
    Serial.println("Attempting MQTT connection...");
    // Create a random client ID
    String clientId = "WemosSensorClient-";
    clientId += String(random(0xffff), HEX);
    // Attempt to connect
    if (MQTTclient.connect(clientId.c_str())) {
      Serial.println("MQTT connected");
    } else {
      Serial.print("failed, rc=");
      Serial.print(MQTTclient.state());
      Serial.println(" try again in 5 seconds");
      // Wait 5 seconds before retrying
      delay(5000);
    }
  }
  //connected, subscribe to topic
  mqttSubscribeTopic(mqttTopicCatcherArm);
}

void mqttHandleClientMessage() {
  // to call in the loop of the arduino
  if (!MQTTclient.connected()) {
    // LED on to  indicate connection problem
    digitalWrite(LED_BUILTIN, LOW);
    mqttConnect();
  } else {
    // LED off all is ok
    digitalWrite(LED_BUILTIN, HIGH);
  }
  MQTTclient.loop(); //initialize the loop to get messages from client
}

void setup() {
  pinMode(BUILTIN_LED, OUTPUT);     // Initialize the BUILTIN_LED pin as an output
  Serial.begin(9600);
  Serial.println("\nInit setup...");
  
  setup_wifi();
  setup_httpServer();
  setup_servos();
  //MQTT
  setupMqttClient();

  Serial.println("RoboticArm ready!");
}

void loop() {

  mqttHandleClientMessage();
  handleClientRequestServer();

  //catch object
  if(catch_object){
    Serial.println("catching...");
    //open catcher
    move_nicely_servo(servoD4, servoD4_min);
    //move arm to position to catch
    //close catcher
    move_nicely_servo(servoD4, servoD4_max);
    //move up
    catch_object = 0;
  }
}
